#!/bin/bash

# This script will try to build and deploy the site.
#
# Has external dependencies to echo, sed, git and yq
#
# Public/ is used as a sub repository and holds the site build.
#
# The production folder will be wiped on every build. Because of this,
# the folder for the build site's git repository cannot be situated in the
# public folder. Instead it is placed in the project root. This script then
# creates a pointer to the configuration directory in the pubic/.git.
#
# The script takes two arguments:
# - One of these environments: 'production', 'staging'.
# - A base domain for staging, for instance 'accodeing.dev', which will
#   be the base domain if the site should be deployed to staging. This
#   argument is only needed if publishing to staging.

# Exit if some variable is uninitialised
set -o nounset

# Exit on error
set -o errexit

############
# FUNCTIONS
############

echo_red () {
  echo -e "\e[31m$1\e[0m"
}

echo_green () {
  echo -e "\e[32m$1\e[0m"
}

echo_yellow () {
  echo -e "\e[33m$1\e[0m"
}

check_config () {
  local CONFIG=$1

  if [ ! -f "$CONFIG" ]; then
    echo "Error: Could not find $CONFIG!"
    exit 2
  fi
}

read_site_name () {
  local CONFIG=$1
  local KEY='name: '

  LINE=$(grep $KEY $CONFIG)
  echo ${LINE/$KEY/}
}

disallow_robots () {
  echo 'Disallowing robots in robots.txt'

  echo -e "User-agent: *\nDisallow: /\n" > robots.txt
}

deploy () {
  echo_yellow "\nDeploying!\n"
  git add --all
  git commit -m "$COMMIT_MESSAGE"
  git push origin master

  echo_green "\nDeployment done!"
}

add_dummy_counter () {
  local CONFIG=$1

  echo 'Adding dummy counter to force deployment'

  yq e ".dummy_counter = \"$DUMMY_COUNTER\"" -i $CONFIG
}

adapt_config_for_staging () {
  local CONFIG=$1
  local SITE_NAME=$2
  local DOMAIN=$3

  echo 'Adapting config for staging'
  yq e ".name = \"$SITE_NAME-staging\"" -i $CONFIG
  yq e ".domains =  [\"$DOMAIN\"]" -i $CONFIG

  add_dummy_counter $CONFIG
}

please_confirm () {
  local SITE_NAME=$1
  local ENVIRONMENT=$2

  read -p "Deploy $SITE_NAME to $ENVIRONMENT? [Y/n] " -n 1
  echo # new line
  if [[ $REPLY =~ ^[Nn]$ ]]; then
    echo_red "Ok, I'll abort!"
    exit
  fi
}

build_site () {
  echo_yellow "Building site, this may take a while...\n"
  yarn build
  exit_on_fail $? 'Build failed!'
  echo # new line
}

check_branch () {
  local CURRENT_BRANCH=$1
  local EXPECTED_BRANCH=$2
  local ENVIRONMENT=$3

  if [ "$CURRENT_BRANCH" != "$EXPECTED_BRANCH" ]; then
    echo "You should be on $EXPECTED_BRANCH branch to deploy to $ENVIRONMENT!"
    exit 1
  fi
}

exit_on_fail () {
  local EXIT_VALUE=$1
  local ERROR_MESSAGE=$2


  if [ "$EXIT_VALUE" != "0" ]; then
    echo_red "Error: $ERROR_MESSAGE"

    echo # new line
    echo_yellow "Trying to cleanup by removing git tag $TAG"
    git tag -d $TAG
    git push --delete origin $TAG

    exit 1
  fi
}

setup_git_deployment () {
  local BUILD_DIR=$1
  local GIT_DIR=$2

  local SOURCE_REPO_URL=$(git config --get remote.origin.url)

  # Origin url will be same as source repo but without the `.src` part.
  # Ex: git@.../example.com.src.git -> git@.../example.com.git
  local ORIGIN_URL=$(echo $SOURCE_REPO_URL | sed "s/\(^.*\)src.git$/\1git/")

  if [ -d "$BUILD_DIR" ]; then
    rm -rf $BUILD_DIR/* # Clean build dir to allow fetching from origin
  else
    mkdir $BUILD_DIR
  fi

  if [ ! -d "$GIT_DIR" ]; then
    echo "Git deployment directory was missing so I'll create $GIT_DIR"

    git init $BUILD_DIR
    mv $BUILD_DIR/.git $GIT_DIR
    create_git_pointer $GIT_DIR $BUILD_DIR

    echo "Setting $ORIGIN_URL as origin"
    git --git-dir $BUILD_DIR/.git remote add origin $ORIGIN_URL
  else
    create_git_pointer $GIT_DIR $BUILD_DIR
  fi

  # Check if remote is initialized and fetch data if so
  if git --git-dir $BUILD_DIR/.git ls-remote --heads --exit-code $ORIGIN_URL 'master'
  then
    cd $BUILD_DIR # Better to change directory when manipulating files...

    echo 'Pull content from master on origin to get up to date.'
    git pull origin master

    echo 'Removing everything from previous build and stage changes to get a fresh build on next commit'
    git rm -rf --quiet . && git add --all

    cd ../
  else
    echo 'No content to pull from origin.'
  fi
}

create_git_pointer () {
  local GIT_DIR=$1
  local BUILD_DIR=$2

  echo "gitdir: ../$GIT_DIR" > $BUILD_DIR/.git
}

############
# GLOBALS
############

ENVIRONMENT=$1
PRODUCTION='production'
STAGING='staging'
CURRENT_BRANCH=$(git symbolic-ref --short HEAD)
GIT_DEPLOYMENT_CONFIG_DIR='.git_deployment'
BUILD_DIR='public/'
CONFIG='config.yaml'
DUMMY_COUNTER=$(date +%s) # Something unique

###############
# START SCRIPT
###############

if [ "$ENVIRONMENT" != "$PRODUCTION" ] && [ "$ENVIRONMENT" != "$STAGING" ]; then
  echo "Error! Unknown environment: $ENVIRONMENT"
  exit 1
fi

if [ "$ENVIRONMENT" == "$PRODUCTION" ]; then
  EXPECTED_BRANCH='master'
elif [ "$ENVIRONMENT" == "$STAGING" ]; then
  EXPECTED_BRANCH='staging'
fi

check_branch $CURRENT_BRANCH $EXPECTED_BRANCH $ENVIRONMENT

SITE_NAME=$(read_site_name $CONFIG)

if [ "$ENVIRONMENT" == "$STAGING" ]; then
  STAGING_BASE_DOMAIN=$2
  STAGING_DOMAIN=$SITE_NAME.$STAGING_BASE_DOMAIN
  echo_yellow "\nStaging domain is $STAGING_DOMAIN\n"
fi

please_confirm $SITE_NAME $ENVIRONMENT

check_config $CONFIG

setup_git_deployment $BUILD_DIR $GIT_DEPLOYMENT_CONFIG_DIR

if [ "$ENVIRONMENT" == "$PRODUCTION" ]; then
  # Semver makes no sense for webpages. Instead, this method tried to build
  # a version number automatically from git data. Commit count will hopefully
  # increase over time and therefore give an increasing version number for each
  # release. The SHA reflects what commit is being used (if the commit count
  # should not have increased) and the timestamp is to make it truly unique.
  # You might want to deploy the same code version multiple for several reasons.
  COMMIT_COUNT=$(git rev-list --count HEAD)
  SHA=$(git rev-parse --short HEAD)
  TAG="v$COMMIT_COUNT-$SHA-$DUMMY_COUNTER"

  git tag -a $TAG -m "Commit count $COMMIT_COUNT, SHA $SHA"
  exit_on_fail $? 'Could not create tag!'

  git push origin $TAG
  exit_on_fail $? "Could not push $TAG to origin"

  COMMIT_MESSAGE="production: $TAG"
fi

if [ "$ENVIRONMENT" == "$STAGING" ]; then
  read -p 'Please give me a commit message to use: ' MESSAGE
  echo # new line
  COMMIT_MESSAGE="staging: $MESSAGE"
fi

build_site

create_git_pointer $GIT_DEPLOYMENT_CONFIG_DIR $BUILD_DIR # Add pointer after build since build wipes build directory

echo_green "Build complete!"

echo -e "\nCopying config"
cp $CONFIG $BUILD_DIR/

cd $BUILD_DIR

if [ "$ENVIRONMENT" == "$PRODUCTION" ]; then
  add_dummy_counter $CONFIG
elif [ "$ENVIRONMENT" == "$STAGING" ]; then
  adapt_config_for_staging $CONFIG $SITE_NAME $STAGING_DOMAIN
  disallow_robots
fi

deploy

cd ../ # Go back to project directory
